# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the command line for the export with ggml."""

from ...exporters.ggml.convert import main_export


class GgmlExportCommand:
    def __init__(self, args):
        self.args = args

    def run(self):
        main_export(
            model_name_or_path=self.args.model,
            output=self.args.output,
            task=self.args.task,
            fp16=self.args.fp16,
        )
